"""

The module datahub is the data manager in GeospaceLab. The datahub has three core class-based components:

    - :class:`DataHub <geospacelab.datahub.DataHub>` manages a set of datasets docked or added to the datahub.
    - :class:`Dataset <geospacelab.datahub.DatasetModel>` manages a set of variables loaded from a data source.
    - :class:`Variable <geospacelab.datahub.VariableModel>` records the value, error, and various attributes \
    (e.g., name, label, unit, depends, ndim, ...) of a variable.

"""

# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLAB (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

# __all__ = ['DataHub', 'DatasetModel', 'VariableModel',]

import importlib
import datetime
import pathlib

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pybasic as pybasic
from geospacelab.datahub.__metadata_base__ \
    import DatabaseModel, MetadataModel, FacilityModel, SiteModel, ProductModel, InstrumentModel
from geospacelab.datahub.__dataset_base__ import DatasetBase, DatasetUser, DatasetSourced
from geospacelab.datahub.__variable_base__ import Visual
from geospacelab.datahub.__variable_base__ import VariableBase as VariableModel
from geospacelab import preferences as pfr


def example():
    dt_fr = datetime.datetime.strptime('20210309' + '0000', '%Y%m%d%H%M')
    dt_to = datetime.datetime.strptime('20210309' + '2359', '%Y%m%d%H%M')
    database_name = 'madrigal'
    facility_name = 'eiscat'

    dh = DataHub(dt_fr, dt_to)
    dh.list_sourced_datasets()
    ds_1 = dh.dock(datasource_contents=[database_name, facility_name],
                          site='UHF', antenna='UHF', modulation='ant', data_file_type='eiscat-hdf5')
    ds_1.load_data()
    var_1 = dh.assign_variable('n_e')
    var_2 = dh.assign_variable('T_i')


def create_datahub(dt_fr, dt_to, visual='off', datahub_class=None, **kwargs):
    """
    Create a datahub object.

    :param dt_fr: The starting time.
    :type dt_fr: datetime.datetime
    :param dt_to: The stopping time.
    :type dt_to: datetime.datetime
    :param visual: If "on", a Visual object is aggregated to the Variable object.
    :type visual: {'off', 'on'}, default: 'off'
    :param datahub_class: If ``None``, create a datahub object based on the
        default :class:`DataHub <geospacelab.datahub.DataHub>` class.
    :type datahub_class: DataHub or its subclass
    :param kwargs: Other optional keyword arguments as inputs to DataHub.
    :return: dh
    :rtype: DataHub object

    :Example:
    >>> import geospacelab.datahub as datahub
    >>> import datetime
    >>> dt_fr = datetime.datetime.strptime('20210309' + '0000', '%Y%m%d%H%M')
    >>> dt_to = datetime.datetime.strptime('20210309' + '2359', '%Y%m%d%H%M')
    >>> dh = datahub.create_datahub(dt_fr, dt_to)

    :seealso:: :class:`DataHub <geospacelab.datahub.DataHub>`
    """

    if datahub_class is None:
        datahub_class = DataHub

    dh = datahub_class(dt_fr=dt_fr, dt_to=dt_to, visual=visual, **kwargs)

    return dh


class DataHub(object):
    """
    The class DataHub manage a set of datasets from various data sources.

    :ivar datetime.datetime dt_fr: The starting time.
    :ivar datetime.datetime dt_to: The ending time.
    :ivar str, {'off', 'on'} visual: If "on", a Visual object will be aggregated to the Variable object.
    :ivar dict datasets:  A *dict* stores the datasets added (:meth:`add_dataset`)
        or docked (:meth:`dock`) to the datahub.
    :ivar dict variables: A *dict* stores the variables assigned from their aggregated datasets.
        Typically used for the dashboards or the I/O configuration.

    **Usage**:

        - Create a DataHub object:

        :Example: Import the datahub module and create a DataHub object

            >>> import geospacelab.datahub as datahub
            >>> import datetime
            >>> dt_fr = datetime.datetime.strptime('20210309' + '0000', '%Y%m%d%H%M')
            >>> dt_to = datetime.datetime.strptime('20210309' + '2359', '%Y%m%d%H%M')
            >>> dh = datahub.DataHub(dt_fr, dt_to)

        :seealso:: :func:`create_datahub <geospacelab.datahub.create_datahub>`

        - Dock a built-in dataset:

        :Example: Dock a EISCAT dataset

            >>> database_name = 'madrigal'      # built-in sourced database name
            >>> facility_name = 'eiscat'
            >>> site = 'UHF'      # facility attributes required, check from the eiscat schedule page
            >>> antenna = 'UHF'
            >>> modulation = 'ant'
            >>> ds_1 = dh.dock(datasource_contents=[database_name, facility_name], site=site, antenna=antenna, modulation=modulation, data_file_type='eiscat-hdf5')

        :seealso:: :meth:`dock`

        - How to know ``datasource_contents`` and required inputs?

        :Example: List the buit-in data sources

            >>> dh.list_sourced_datasets()

        :seealso:: :meth:`list_sourced_datasets`

    """

    __dataset_model__ = DatasetBase
    __variable_model__ = VariableModel

    def __init__(self, dt_fr=None, dt_to=None, visual='off', **kwargs):
        """
            :param dt_fr: The starting time.
            :type dt_fr: datetime.datetime
            :param dt_to: The stopping time.
            :type dt_to: datetime.datetime
            :param visual: If "on", a Visual object is aggregated to the Variable object.
            :type visual: {'off', 'on'}, default: 'off'
            :param kwargs: other keyword arguments forwarded to the inherited class.
        """

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.visual = visual
        self.datasets = {}
        self.variables = {}
        self._current_dataset = None

        super().__init__(**kwargs)

    def dock(self, datasource_contents, **kwargs) -> DatasetSourced:
        """Dock a built-in or registered dataset.

        :param datasource_contents: the contents that required for docking a sourced dataset.
            To look up the sourced dataset and the associated contents, call :func:`~geospacelab.datahub.DataHub.list_sourced_datasets()`.
        :type datasource_contents: list
        :param dt_fr: starting time, optional, use datahub.dt_fr if not specified.
        :type dt_fr: datetime.datetime
        :param dt_to: stopping time, optional, use datahub.dt_to if not specified.
        :type dt_to: datetime.datetime
        :param visual: variable attribute, use datahub.visual if not specified.
        :type visual: str
        :return: ``dataset``
        :rtype: :class:`Dataset <geospacelab.datahub.DatasetModel>` object

        :seealso:: :meth:`add_dataset`

        :note:: The difference between the methods :meth:`dock` and :meth:`add_dataset` is that :meth:`dock` adds a built-in data
            source, while :meth:`add_dataset` add a temporary or user-defined dataset, which is not initially included in the
            package.

        -------------------

        """
        kwargs.setdefault('dt_fr', self.dt_fr)
        kwargs.setdefault('dt_to', self.dt_to)
        kwargs.setdefault('visual', self.visual)
        append = True

        module_keys = [pfr.package_name, 'datahub', 'sources']
        module_keys.extend(datasource_contents)
        try:
            module = importlib.import_module('.'.join(module_keys))
            dataset = getattr(module, 'Dataset')(**kwargs)
            dataset.kind = 'sourced'
        except Exception as error:
            error_str = str(error)
            print(error_str.replace(pfr.package_name + '.datahub.sources.', ''))
            mylog.StreamLogger.error(
                'The data source cannot be docked. \n'
                + 'Check the built-in sourced data using the method: "list_sourced_dataset". \n'
                + 'Or, add a temporary or user-defined dataset by the method of "add_dataset". '
            )
            return

        # if dataset.label() in [ds.label() for ds in self.datasets]:
        #     mylog.simpleinfo.info('The same dataset has been docked!')
        #    append = False

        if append:
            self._append_dataset(dataset)

        return dataset

    def add_dataset(self, *args, kind='temporary', dataset_class=None, **kwargs) -> __dataset_model__:
        """Add one or more datasets, which can be a "temporary" or "user-defined" dataset.

        :param args: A list of the datasets.
        :type args: list(dataset)
        :param kind: The type of a dataset. If temporary, a new dataset will be created from the DatasetModel.
        :type kind: {'temporary', 'user-defined'}, default: 'temporary'
        :param dataset_class: If None, the default class is DatasetModel. Used when ``kind='temporary'``.
        :type dataset_class: DatasetModel or its subclass
        :param kwargs: Other keyword arguments forwarded to ``dataset_class``
        :return: None

        :seealso:: :meth:`dock`

        """
        kwargs.setdefault('dt_fr', self.dt_fr)
        kwargs.setdefault('dt_to', self.dt_to)
        kwargs.setdefault('visual', self.visual)
        kwargs.setdefault('datasets', [])

        if dataset_class is None:
            dataset_class = self.__dataset_model__

        if kind == 'temporary':
            kwargs.pop('datasets', None)
            dataset = dataset_class(name='temporary', **kwargs)
            self._append_dataset(dataset)
            return dataset
        elif len(args) == 0 and kind in ["user-defined", "UserDefined", "UD"]:
            kwargs.pop('datasets', None)
            dataset = DatasetUser(name='', **kwargs)
            self._append_dataset(dataset)
            return dataset
        else:
            kind = "user-defined"

        for dataset in args:
            kwargs['datasets'].append(dataset)

        for dataset in kwargs['datasets']:
            kind = 'user-defined'
            if issubclass(dataset.__class__, DatasetUser):
                dataset.kind = kind
            else:
                TypeError('A dataset instance\'s class must be a heritage of DatasetUser!')
            self._append_dataset(dataset)

        return None

    def _append_dataset(self, dataset):
        """Append a dataset to the datahub

        :param dataset: a dataset (Dataset object)
        :return: None
        """

        ind = len(self.datasets.keys())
        name = 'dataset_{:02d}'.format(ind)
        if dataset.name is None:
            dataset.name = name
        self.datasets[ind] = dataset
        self.set_current_dataset(dataset=dataset)

    def set_current_dataset(self, dataset=None, dataset_index=None):
        """
        Set the current dataset.

        :param dataset: A Dataset object.
        :param dataset_index: The index of the dataset in ``.datasets``.
        :type dataset_index: int
        :rtype: None
        """
        if dataset is not None:
            if dataset in self.datasets.values():
                cds = dataset
            else:
                raise ValueError("The input dataset is not aggregated to the datahub!")
        elif dataset_index is not None:
            cds = self.datasets[dataset_index]
        else:
            raise ValueError("Either dataset or dataset_index must be set!")
        self._current_dataset = cds

    def get_current_dataset(self, index=False):
        """
        Get the current dataset.

        :param index: The index of the dataset.
        :type index: bool
        :return: If ``index=False``, dataset object, else dataset_index.
        """

        if index:
            datasets_r = {value: key for key, value in self.datasets.items()}
            res = datasets_r[self._current_dataset]
        else:
            res = self._current_dataset

        return res

    def get_variable(self, var_name, dataset=None, dataset_index=None) -> __variable_model__:
        """To get a variable from the docked or added dataset.

        :param str var_name: the name of the queried variable
        :param DatasetBase object dataset: the dataset storing the queried variable.
        :param int dataset_index: the index of the dataset in datahub.datasets.
            if both dataset or dataset_index are not specified, the function will get the
            variable from the current dataset.
        :return: var
        :rtype: :class:`VariableModel <geospacelab.datahub.VariableModel>` object or None

        :seealso: :meth:`assign_variable`

        :note:: Both :meth:`get_variable` and :meth:`assign_variable` return a variable object assigned from a dataset.
            The former only returns the object, and the latter also assign the variable to the ``DataHub.variables``.
        """

        if dataset is None and dataset_index is None:
            dataset = self.get_current_dataset()

        if dataset is None:
            dataset = self.datasets[dataset_index]
        elif not issubclass(dataset.__class__, DatasetBase):
            raise TypeError('A dataset instance\'s class must be an inheritance of DatasetBase!')

        if dataset.exist(var_name):
            var = dataset[var_name]
        else:
            mylog.StreamLogger.warning(
                f"Cannot find the queried variable in {repr(dataset)}. Return None.")
            var = None

        return var

    def assign_variable(self, var_name, dataset=None, dataset_index=None, add_new=False, **kwargs) -> __variable_model__:
        """Assign a variable to `DataHub.variables` from the docked or added dataset.

        :param var_name: The name of the variable
        :param dataset: The dataset that stores the variable
        :param dataset_index: The index of the dataset in the datahub.datasets.
        :param add_new: if True, add the variable to the specified dataset and assign to the datahub
        :param kwargs: other keywords to configure the attributes of the variable.
        :return: object of :class:`~geospacelab.datahub.VariableModel`

        :seealso:: :meth:`get_variable`

        """
        if dataset is None:
            if dataset_index is None:
                dataset = self.get_current_dataset()  # the current dataset
            else:
                dataset = self.datasets[dataset_index]
        elif not issubclass(dataset.__class__, DatasetBase):
            raise TypeError('A dataset instance\'s class must be a heritage of the class DatasetBase!')

        var = self.get_variable(var_name, dataset=dataset, dataset_index=dataset_index)

        if var is None:
            if add_new:
                var = dataset.add_variable(var_name)
            else:
                raise KeyError('The variable does not exist in the dataset. Set add_new=True, if you want to add.')
        var.config(**kwargs)

        if var not in self.variables.values():
            self._append_variable(var)

        return var

    def _append_variable(self, var):
        """ append a variable to datahub.variables

        :param var: a object of :class:`~geospacelab.datahub.VariableModel`
        :return: None
        """
        ind = len(self.variables.keys())
        self.variables[ind] = var

    @property
    def host_dataset(self):
        return self.datasets[0]

    @staticmethod
    def list_sourced_datasets():
        """List all the bult-in data sources this package

        The list will be printed in the python console in a \"tree\" view.
        """
        this_file_dir = pathlib.Path(__file__).resolve().parent
        data_source_dir = this_file_dir / 'sources'
        sub_dirs = list(data_source_dir.glob("**"))
        data_sources = {}
        for sub_dir in sub_dirs:
            init_file = sub_dir / "__init__.py"
            if not init_file.is_file():
                continue
            module_name = str(sub_dir).replace('/', '.').split(pfr.package_name + '.', 1)[-1]
            try:
                module = importlib.import_module(module_name)
            except:
                continue
            try:
                getattr(module, 'Dataset')
            except AttributeError:
                continue

            contents = module_name.split('sources.')[-1].split('.')
            current_dict = data_sources
            try:
                required_inputs = getattr(module, 'default_attrs_required')
            except AttributeError:
                required_inputs = []

            for ind, content in enumerate(contents):
                content = content.upper()
                current_dict.setdefault(content, {})
                if ind == len(contents) - 1:
                    required_inputs = ['datasource_contents', *required_inputs]
                    current_dict[content] = {
                        'Required inputs when load_mode="AUTO"': required_inputs,
                        'datasource_contents': contents,
                    }
                    continue
                current_dict = current_dict[content]

        pybasic.dict_print_tree(data_sources, full_value=False, dict_repr=False, value_repr=True, max_level=None)

    def list_datasets(self):
        """ List all the datasets that have been docked or added to the datahub

        The list will be printed in the console as a table
        """
        # print header
        mylog.simpleinfo.info("Listing datasets ...")
        mylog.simpleinfo.info('{:^20s}{:60s}'.format('Index', 'Dataset'))
        for ind, dataset in self.datasets.items():
            dataset_label = dataset.label()
            mylog.simpleinfo.info('{:^20d}{:60s}'.format(ind, dataset_label))
        print()

    def list_assigned_variables(self):
        """ List all the assigned variables that have been docked or added to the datahub

        The list will be printed in the console as a table
        """
        # print header
        mylog.simpleinfo.info("Listing the assigned variables ...")

        mylog.simpleinfo.info('{:^20s}{:30s}{:60s}'.format('Index', 'Variable name', 'Dataset'))
        for ind, var in self.variables.items():
            dataset_label = var.dataset.label()
            var_name = var.name
            mylog.simpleinfo.info('{:^20d}{:30s}{:60s}'.format(ind, var_name, dataset_label))
        print()

    def save_to_pickle(self):
        pass

    def save_to_cdf(self):
        pass

    def __repr__(self):
        name = self.__class__.__name__
        r =  f"GeospaceLab {name} object <starting time: {str(self.dt_fr)} and stopping time: {str(self.dt_to)}>"
        return r


if __name__ == "__main__":
    example()



import numpy as np
from geopack import t89,t96,t01,t04
import os.path
import datetime

def init_igrf():
    """
    Initialize the IGRF coefficients and related coefs.
    Should be called once and only once when importing the geopack module.
    """

    global igrf, nmn,mns, nyear,years,yruts

    print('Load IGRF coefficients ...')

    bfn = 'igrf13coeffs.txt'
    locffn = os.path.join(os.path.dirname(__file__), bfn)

    nheader = 3
    with open(locffn, 'r') as file:
        for i in range(nheader):
            next(file)
        header = file.readline().rstrip()
        cols = header.split()
        # first 3 columns are g/h flag, n, m.
        years = np.array([np.int32(j[0:4]) for j in cols[3:]])
        nyear = len(years)
        lines = file.read().splitlines()

    cols = lines[-1].split()
    k = np.int32(cols[1]) + 1
    nmn = np.int32((k + 1) * k * 0.5)
    igrf = np.zeros((nmn, nyear, 2), dtype=float)
    mns = np.empty((nmn, 2), dtype=np.int32)
    l = 0
    for i in range(k):
        for j in range(i + 1):
            mns[l, :] = [i, j]
            l += 1

    for line in lines:
        cols = line.split()
        if cols[0] == 'g':
            i = 0
        else:
            i = 1
        n, m = np.int32(cols[1:3])
        mn = np.int32(n * (n + 1) * 0.5 + m)
        igrf[mn, :, i] = [np.float32(j) for j in cols[3:]]

    # treat the last column
    years[-1] += 5
    igrf[:, -1, :] = igrf[:, -2, :] + igrf[:, -1, :] * 5
    yruts = np.empty(nyear, dtype=float)
    t0_datetime = datetime.datetime(1970,1,1)
    for i in range(nyear):
        yruts[i] = (datetime.datetime(years[i],1,1)-t0_datetime).total_seconds()
    # separate g/h
    igrf = {'g': igrf[:,:,0], 'h': igrf[:,:,1]}


def load_igrf(ut):
    """
    Load the IGRF coefficients for the given time.
    :param ut: ut sec, a float.
    :return: g,h. The IGRF coef at given time.
    """

    # igrf should be initilized already.
    global igrf, nmn,nyear, mns, years, yruts
    try: isinstance(igrf, dict)
    except: init_igrf()

    # locate the two years of interest.
    if ut <= yruts[0]: yridx = 0
    elif ut >= yruts[-1]: yridx = -2
    else:
        yridx = np.argwhere(yruts <= ut)[-1]

    g0 = np.squeeze(igrf['g'][:,yridx])
    h0 = np.squeeze(igrf['h'][:,yridx])
    g1 = np.squeeze(igrf['g'][:,yridx+1])
    h1 = np.squeeze(igrf['h'][:,yridx+1])

    ut0 = yruts[yridx]
    ut1 = yruts[yridx+1]
    f1 = (ut-ut0)/(ut1-ut0)
    f0 = 1-f1

    return g0*f0+g1*f1, h0*f0+h1*f1


def igrf_gsw(xgsw,ygsw,zgsw):
    """
    Calculates components of the main (internal) geomagnetic field in the geocentric solar
    magnetospheric coordinate system, using IAGA international geomagnetic reference model
    coefficients (e.g., http://www.ngdc.noaa.gov/iaga/vmod/igrf.html revised: 22 march, 2005)

    Before the first call of this subroutine, or if the date/time
    was changed, the model coefficients and GEO-GSW rotation matrix elements should be updated
    by calling the subroutine recalc

    Python version by Sheng Tian

    :param xgsw,ygsw,zgsw: cartesian GSW coordinates (in units Re=6371.2 km)
    :return: hxgsw,hygsw,hzgsw. Cartesian GSW components of the main geomagnetic field in nanotesla
    """
    xgsm,ygsm,zgsm = gswgsm(xgsw,ygsw,zgsw, 1)
    bxgsm,bygsm,bzgsm = igrf_gsm(xgsm,ygsm,zgsm)
    return gswgsm(bxgsm,bygsm,bzgsm, -1)


def igrf_gsm(xgsm,ygsm,zgsm):
    """
    Calculates components of the main (internal) geomagnetic field in the geocentric solar
    magnetospheric coordinate system, using IAGA international geomagnetic reference model
    coefficients (e.g., http://www.ngdc.noaa.gov/iaga/vmod/igrf.html revised: 22 march, 2005)

    Before the first call of this subroutine, or if the date/time
    was changed, the model coefficients and GEO-GSM rotation matrix elements should be updated
    by calling the subroutine recalc

    Python version by Sheng Tian

    :param xgsm,ygsm,zgsm: cartesian GSM coordinates (in units Re=6371.2 km)
    :return: hxgsm,hygsm,hzgsm. Cartesian GSM components of the main geomagnetic field in nanotesla
    """

    xgeo,ygeo,zgeo = geogsm(xgsm,ygsm,zgsm, -1)
    r,theta,phi = sphcar(xgeo,ygeo,zgeo, -1)
    br,btheta,bphi = igrf_geo(r,theta,phi)
    bxgeo,bygeo,bzgeo = bspcar(theta,phi,br,btheta,bphi)
    return geogsm(bxgeo,bygeo,bzgeo, 1)


def igrf_geo(r,theta,phi):
    """
    Calculates components of the main (internal) geomagnetic field in the spherical geographic
    (geocentric) coordinate system, using IAGA international geomagnetic reference model
    coefficients  (e.g., http://www.ngdc.noaa.gov/iaga/vmod/igrf.html, revised: 22 march, 2005)

    Before the first call of this subroutine, or if the time was changed,
    the model coefficients should be updated by calling the subroutine recalc

    Python version by Sheng Tian

    :param r: spherical geographic (geocentric) coordinates: radial distance r in units Re=6371.2 km
    :param theta: colatitude theta in radians
    :param phi: longitude phi in radians
    :return: br, btheta, bphi. Spherical components of the main geomagnetic field in nanotesla
        (positive br outward, btheta southward, bphi eastward)
    """


    # common /geopack2/ g(105),h(105),rec(105)
    global g, h, rec

    ct = np.cos(theta)
    st = np.sin(theta)
    minst = 1e-5
    if np.abs(st) < minst: smlst = True
    else: smlst = False

    # In this new version, the optimal value of the parameter nm (maximal order of the spherical
    # harmonic expansion) is not user-prescribed, but calculated inside the subroutine, based
    # on the value of the radial distance r:
    irp3 = np.int64(r+2)
    nm = np.int64(3+30/irp3)
    if nm > 13: nm = 13
    k = nm+1

    # r dependence is encapsulated here.
    a = np.empty(k)
    b = np.empty(k)
    ar = 1/r        # a/r
    a[0] = ar*ar    # a[n] = (a/r)^(n+2).
    b[0] = a[0]     # b[n] = (n+1)(a/r)^(n+2)
    for n in range(1,k):
        a[n] = a[n-1]*ar
        b[n] = a[n]*(n+1)


    # t - short for theta, f - short for phi.
    br,bt,bf = [0.]*3
    d,p = [0.,1]

    # m = 0. P^n,0
    m = 0
    smf,cmf = [0.,1]
    p1,d1,p2,d2 = [p,d,0.,0]
    l0 = 0
    mn = l0
    for n in range(m,k):
        w = g[mn]*cmf+h[mn]*smf
        br += b[n]*w*p1          # p1 is P^n,m.
        bt -= a[n]*w*d1          # d1 is dP^n,m/dt.
        xk = rec[mn]
        # Eq 16c and its derivative on theta.
        d0 = ct*d1-st*p1-xk*d2   # dP^n,m/dt = ct*dP^n-1,m/dt - st*P_n-1,m - K^n,m*dP^n-2,m/dt
        p0 = ct*p1-xk*p2         # P^n,m = ct*P^n-1,m - K^n,m*P^n-2,m
        d2,p2,d1 = [d1,p1,d0]
        p1 = p0
        mn += n+1

    # Eq 16b and its derivative on theta.
    d = st*d+ct*p   # dP^m,m/dt = st*dP^m-1,m-1/dt + ct*P^m-1,m-1
    p = st*p        # P^m,m = st*P^m-1,m-1

    # Similarly for P^n,m
    l0 = 0
    for m in range(1,k):        # sum over m
        smf = np.sin(m*phi)     # sin(m*phi)
        cmf = np.cos(m*phi)     # cos(m*phi)
        p1,d1,p2,d2 = [p,d,0.,0]
        tbf = 0.
        l0 += m+1
        mn = l0
        for n in range(m,k):    # sum over n
            w=g[mn]*cmf+h[mn]*smf   # [g^n,m*cos(m*phi)+h^n,m*sin(m*phi)]
            br += b[n]*w*p1
            bt -= a[n]*w*d1
            tp = p1
            if smlst: tp = d1
            tbf += a[n]*(g[mn]*smf-h[mn]*cmf)*tp
            xk = rec[mn]
            d0 = ct*d1-st*p1-xk*d2   # dP^n,m/dt = ct*dP^n-1,m/dt - st*P_n-1,m - K^n,m*dP^n-2,m/dt
            p0 = ct*p1-xk*p2         # P^n,m = ct*P^n-1,m - K^n,m*P^n-2,m
            d2,p2,d1 = [d1,p1,d0]
            p1=p0
            mn += n+1

        d = st*d+ct*p
        p = st*p

        # update B_phi.
        tbf *= m
        bf += tbf

    if smlst:
        if ct < 0.: bf = -bf
    else: bf /= st

    return br,bt,bf




def dip(xgsm,ygsm,zgsm):
    """
    Calculates gsm components of a geodipole field with the dipole moment
    corresponding to the epoch, specified by calling subroutine recalc (should be
    invoked before the first use of this one and in case the date/time was changed).

    :param xgsm,ygsm,zgsm: GSM coordinates in Re (1 Re = 6371.2 km)
    :return: bxgsm,bygsm,gzgsm. Field components in gsm system, in nanotesla.

    Last modification: May 4, 2005.
    Author: N. A. Tsyganenko
    """

    # common /geopack1/ aaa(10),sps,cps,bbb(23)
    # common /geopack2/ g(105),h(105),rec(105)
    global aaa, sps,cps, bbb, g, h, rec

    dipmom = np.sqrt(g[1]**2+g[2]**2+h[2]**2)

    p = xgsm**2
    u = zgsm**2
    v = 3*zgsm*xgsm
    t = ygsm**2
    q = dipmom/np.sqrt(p+t+u)**5

    bxgsm = q*((t+u-2.*p)*sps-v*cps)
    bygsm = -3.*ygsm*q*(xgsm*sps+zgsm*cps)
    bzgsm = q*((p+t-2.*u)*cps-v*sps)

    return bxgsm,bygsm,bzgsm


def dip_gsw(xgsw,ygsw,zgsw):
    """
    Calculates gsm components of a geodipole field with the dipole moment
    corresponding to the epoch, specified by calling subroutine recalc (should be
    invoked before the first use of this one and in case the date/time was changed).

    :param xgsw,ygsw,zgsw: GSW coordinates in Re (1 Re = 6371.2 km)
    :return: bxgsm,bygsm,gzgsm. Field components in gsm system, in nanotesla.

    Author: Sheng Tian
    """
    xgsm,ygsm,zgsm = gswgsm(xgsw,ygsw,zgsw, 1)
    bxgsm,bygsm,bzgsm = dip(xgsm,ygsm,zgsm)
    return gswgsm(bxgsm,bygsm,bzgsm, -1)


def recalc(ut, vxgse=-400,vygse=0,vzgse=0):
    """
    1. Prepares elements of rotation matrices for transformations of vectors between
        several coordinate systems, most frequently used in space physics.
    2. Prepares coefficients used in the calculation of the main geomagnetic field (igrf model)

    This subroutine should be invoked before using the following subroutines:
        igrf_geo, igrf_gsm, dip, geomag, geogsm, magsm, smgsm, gsmgse, geigeo.
    There is no need to repeatedly invoke recalc, if multiple calculations are made for the same date and time.

    :param ut: Universal time in second.
    :param v[xyz]gse: The solar wind velocity expressed in GSE.
    :return: psi. Dipole tilt angle in radian.

    Python version by Sheng Tian
    """


    # The common block /geopack1/ contains elements of the rotation matrices and other
    # parameters related to the coordinate transformations performed by this package
    # common /geopack1/ st0,ct0,sl0,cl0,ctcl,stcl,ctsl,stsl,sfi,cfi,sps,
    # cps,shi,chi,hi,psi,xmut,a11,a21,a31,a12,a22,a32,a13,a23,a33,ds3,
    # cgst,sgst,ba(6)

    # st0/ct0 - sin/cos of teta0 (colat in geo?). sl0/cl0 - sin/cos of lambda0 (longitude in geo?).
    # ctcl/stcl - . ctsl/stsl - . geo x and z?
    # sfi/cfi/xmut - rotate angle between mag and sm and its sin/cos.
    # sps/cps/psi - tilt angle and its sin/cos.
    # shi/chi/hi - rotate angle between gse to gsm and its sin/cos.
    # a[11,...,33] - matrix converts geo to gsm.
    # cgst/sgst - cos/sin of gst.
    # ds3.
    # ba(6).
    global st0,ct0,sl0,cl0,ctcl,stcl,ctsl,stsl,sfi,cfi,sps,cps, \
        shi,chi,hi,psi,xmut,ds3,cgst,sgst,ba, \
        a11,a21,a31,a12,a22,a32,a13,a23,a33, \
        e11,e21,e31,e12,e22,e32,e13,e23,e33

    # The common block /geopack2/ contains coefficients of the IGRF field model, calculated
    # for a given year and day from their standard epoch values. the array rec contains
    # coefficients used in the recursion relations for legendre associate polynomials.
    # common /geopack2/ g(105),h(105),rec(105)
    global g,h,rec


    # Compute the m,n related coefficients (following the notation in Davis 2004):
    # 1. The Schmidt quasi-normalization: S_n,m, which normalizes the associated Legendre polynomials P_n^m
    #    to the Guassian normalized associated Legendre polynomials P^n,m.
    #
    #    Since g_n^m * P_n^m should be constant, mutiplying S_n,m to g_n^m is equivalently converting P_n^m to P^n,m.
    #    The benefit of doing so is that P^n,m follows simple recursive form, c.f. Eq (16 a-c) and Eq(17 a-b).
    # 2. rec[mn], which used in the recursion relation.
    k = 14
    nmn = np.int32((k+1)*k/2)

    # rec[mn].
    rec = np.empty(nmn,dtype=float)
    mn = 0
    for n in range(k):                  # K^1,m = 0, Eq (17a), automatically done.
        n2 = 2*n+1
        n2 = n2*(n2-2)
        for m in range(n+1):
            rec[mn] = (n-m)*(n+m)/n2    # K^n,m = (n-m)(n+m)/(2n+1)(2n-1), Eq (17b)
            mn += 1

    # coefficients for a given time, g_n^m(t), h_n^m(t)
    g,h = load_igrf(ut)

    # now multiply them by schmidt normalization factors:
    s = 1.                      # S_0,0 = 1, Eq (18a)
    mn = 0
    for n in range(1,k):
        mn += 1                 # skip m=0.
        s *= (2*n-1)/n
        g[mn] *= s              # S_n,0 = S_n-1,0 * (2n-1)/n, Eq (18b)
        h[mn] *= s
        p = s
        for m in range(1,n+1):
            if m == 1: aa = 2   # aa = delta_m,1
            else: aa = 1
            p *= np.sqrt(aa*(n-m+1)/(n+m))
            mn += 1
            g[mn] *= p          # S_n,m = S_n,m-1 * sqrt(aa(n-m+1)/(n+m)), Eq (18c)
            h[mn] *= p          # now g/h are actually g^n,m, Eq (14 a-b)

    g10=-g[1]
    g11=-g[2]
    h11=-h[2]

    # Now calculate the components of the unit vector ezmag in geo coord.system:
    # sin(teta0)*cos(lambda0), sin(teta0)*sin(lambda0), and cos(teta0)
    #       st0 * cl0                st0 * sl0                ct0
    sq=g11**2+h11**2
    sqq=np.sqrt(sq)
    sqr=np.sqrt(g10**2+sq)
    sl0= h11/sqq
    cl0= g11/sqq
    st0= sqq/sqr
    ct0= g10/sqr
    stcl=st0*cl0
    stsl=st0*sl0
    ctsl=ct0*sl0
    ctcl=ct0*cl0


    gst,slong,srasn,sdec,obliq = sun(ut)

    # All vectors are expressed in GEI.

    # xgse_[xyz] (s[123]) are the components of the unit vector exgsm=exgse in GEI,
    # pointing from the earth's center to the sun:
    xgse_x=np.cos(srasn)*np.cos(sdec)
    xgse_y=np.sin(srasn)*np.cos(sdec)
    xgse_z=np.sin(sdec)

    # zgse_[xyz] (dz[123]) in GEI has the components (0,-sin(delta),cos(delta)) = (0.,-0.397823,0.917462);
    # Here delta = 23.44214 deg for the epoch 1978 (see the book by gurevich or other astronomical handbooks).
    # Here the most accurate time-dependent formula is used:
    zgse_x=0.
    zgse_y=-np.sin(obliq)
    zgse_z= np.cos(obliq)

    # ygse_[xyz] (dy[123]) = zgse_[xyz] x xgsm_[xyz] in GEI:
    ygse_x=zgse_y*xgse_z-zgse_z*xgse_y
    ygse_y=zgse_z*xgse_x-zgse_x*xgse_z
    ygse_z=zgse_x*xgse_y-zgse_y*xgse_x

    # zsm_[xyz] (dip[123]) are the components of the unit vector zsm=zmag in GEI:
    cgst=np.cos(gst)
    sgst=np.sin(gst)
    zsm_x=stcl*cgst-stsl*sgst
    zsm_y=stcl*sgst+stsl*cgst
    zsm_z=ct0

    # xgsw_[xyz] (x[123]) in GEI.
    v1 = -1/np.sqrt(vxgse*vxgse+vygse*vygse+vzgse*vzgse)
    xgsw_x = (vxgse*xgse_x + vygse*ygse_x + vzgse*zgse_x)*v1
    xgsw_y = (vxgse*xgse_y + vygse*ygse_y + vzgse*zgse_y)*v1
    xgsw_z = (vxgse*xgse_z + vygse*ygse_z + vzgse*zgse_z)*v1

    # ygsw (y[123]) = zsm x xgsw in GEI.
    ygsw_x=zsm_y*xgsw_z-zsm_z*xgsw_y
    ygsw_y=zsm_z*xgsw_x-zsm_x*xgsw_z
    ygsw_z=zsm_x*xgsw_y-zsm_y*xgsw_x
    y=np.sqrt(ygsw_x*ygsw_x+ygsw_y*ygsw_y+ygsw_z*ygsw_z)
    ygsw_x=ygsw_x/y
    ygsw_y=ygsw_y/y
    ygsw_z=ygsw_z/y

    # zgsw (z[123]) = xgsw x ygsw in GEI.
    zgsw_x = xgsw_y*ygsw_z-xgsw_z*ygsw_y
    zgsw_y = xgsw_z*ygsw_x-xgsw_x*ygsw_z
    zgsw_z = xgsw_x*ygsw_y-xgsw_y*ygsw_x


    # xgsm = xgse in GEI.
    xgsm_x,xgsm_y,xgsm_z = xgse_x,xgse_y,xgse_z

    # ygsm = zsm x xgsm in GEI.
    ygsm_x = zsm_y*xgsm_z - zsm_z*xgsm_y
    ygsm_y = zsm_z*xgsm_x - zsm_x*xgsm_z
    ygsm_z = zsm_x*xgsm_y - zsm_y*xgsm_x
    y=np.sqrt(ygsm_x*ygsm_x+ygsm_y*ygsm_y+ygsm_z*ygsm_z)
    ygsm_x = ygsm_x/y
    ygsm_y = ygsm_y/y
    ygsm_z = ygsm_z/y

    # ezgsm = exgsm x eygsm in GEI.
    zgsm_x = xgse_y*ygsm_z-xgse_z*ygsm_y
    zgsm_y = xgse_z*ygsm_x-xgse_x*ygsm_z
    zgsm_z = xgse_x*ygsm_y-xgse_y*ygsm_x

    # The elements of the matrix gse to gsm are the scalar products:
    # chi=em22=(eygsm,eygse), shi=em23=(eygsm,ezgse), em32=(ezgsm,eygse)=-em23, and em33=(ezgsm,ezgse)=em22
    chi = ygsm_x*ygse_x + ygsm_y*ygse_y + ygsm_z*ygse_z
    shi = ygsm_x*zgse_x + ygsm_y*zgse_y + ygsm_z*zgse_z
    hi = np.arcsin(shi)

    # elements of the matrix gsm to gsw are the scalar products:
    # e11 = (exgsm, exgsw) e12 = (exgsm, eygsw) e13 = (exgsm, ezgsw)
    # e21 = (eygsm, exgsw) e22 = (eygsm, eygsw) e23 = (eygsm, ezgsw)
    # e31 = (ezgsm, exgsw) e32 = (ezgsm, eygsw) e33 = (ezgsm, ezgsw)
    e11 = xgsm_x*xgsw_x + xgsm_y*xgsw_y + xgsm_z*xgsw_z
    e12 = xgsm_x*ygsw_x + xgsm_y*ygsw_y + xgsm_z*ygsw_z
    e13 = xgsm_x*zgsw_x + xgsm_y*zgsw_y + xgsm_z*zgsw_z
    e21 = ygsm_x*xgsw_x + ygsm_y*xgsw_y + ygsm_z*xgsw_z
    e22 = ygsm_x*ygsw_x + ygsm_y*ygsw_y + ygsm_z*ygsw_z
    e23 = ygsm_x*zgsw_x + ygsm_y*zgsw_y + ygsm_z*zgsw_z
    e31 = zgsm_x*xgsw_x + zgsm_y*xgsw_y + zgsm_z*xgsw_z
    e32 = zgsm_x*ygsw_x + zgsm_y*ygsw_y + zgsm_z*ygsw_z
    e33 = zgsm_x*zgsw_x + zgsm_y*zgsw_y + zgsm_z*zgsw_z


    # Tilt angle: psi=arcsin(ezsm dot exgsm)
    sps=zsm_x*xgse_x+zsm_y*xgse_y+zsm_z*xgse_z
    cps=np.sqrt(1.-sps**2)
    psi=np.arcsin(sps)

    # The elements of the matrix mag to sm are the scalar products:
    # cfi=gm22=(eysm,eymag), sfi=gm23=(eysm,exmag); They can be derived as follows:
    # In geo the vectors exmag and eymag have the components (ct0*cl0,ct0*sl0,-st0) and (-sl0,cl0,0), respectively.

    # Hence, in gei the components are:
    #     exmag:    ct0*cl0*cos(gst)-ct0*sl0*sin(gst)
    #               ct0*cl0*sin(gst)+ct0*sl0*cos(gst)
    #               -st0
    #     eymag:    -sl0*cos(gst)-cl0*sin(gst)
    #               -sl0*sin(gst)+cl0*cos(gst)
    #                0
    # The components of eysm in gei were found above as ysm_in_geix, ysm_in_geiy, and ysm_in_geiz;
    # Now we only have to combine the quantities into scalar products:
    xmag_x= ct0*(cl0*cgst-sl0*sgst)
    xmag_y= ct0*(cl0*sgst+sl0*cgst)
    xmag_z=-st0
    ymag_x=-(sl0*cgst+cl0*sgst)
    ymag_y=-(sl0*sgst-cl0*cgst)
    cfi=ygsm_x*ymag_x+ygsm_y*ymag_y
    sfi=ygsm_x*xmag_x+ygsm_y*xmag_y+ygsm_z*xmag_z

    xmut=(np.arctan2(sfi,cfi)+3.1415926536)*3.8197186342

    # The elements of the matrix geo to gsm are the scalar products:
    # a11=(exgeo,exgsm), a12=(eygeo,exgsm), a13=(ezgeo,exgsm),
    # a21=(exgeo,eygsm), a22=(eygeo,eygsm), a23=(ezgeo,eygsm),
    # a31=(exgeo,ezgsm), a32=(eygeo,ezgsm), a33=(ezgeo,ezgsm),
    # All the unit vectors in brackets are already defined in gei:
    # xgeo=(cgst,sgst,0), ygeo=(-sgst,cgst,0), zgeo=(0,0,1)
    # and therefore:
    a11= xgsm_x*cgst+xgse_y*sgst
    a12=-xgsm_x*sgst+xgse_y*cgst
    a13= xgsm_z
    a21= ygsm_x *cgst+ygsm_y *sgst
    a22=-ygsm_x *sgst+ygsm_y *cgst
    a23= ygsm_z
    a31= zgsm_x*cgst+zgsm_y*sgst
    a32=-zgsm_x*sgst+zgsm_y*cgst
    a33= zgsm_z

    return psi

def sun(ut):
    """
    Calculates four quantities necessary for coordinate transformations
    which depend on sun position (and, hence, on universal time and season)
    Based on http://aa.usno.navy.mil/faq/docs/SunApprox.php and http://aa.usno.navy.mil/faq/docs/GAST.php

    :param ut: ut sec, can be array.
    :return: gst,slong,srasn,sdec. gst - greenwich mean sidereal time, slong - longitude along ecliptic
        srasn - right ascension,  sdec - declination of the sun (radians)
        obliq - mean oblique of the ecliptic (radian)

    Python version by Sheng Tian
    """

    twopi = 2*np.pi
    jd2000 = 2451545.0
    # convert to Julian date.
    t0_jd = 2440587.5       # in day, 0 of Julian day.
    secofdaygsw_x = 1./86400    # 1/sec of day.
    t_jd = ut*secofdaygsw_x+t0_jd

    # d = mjd - mj2000.
    d = t_jd-jd2000
    d = np.squeeze(d)

    # mean obliquity of the ecliptic, e.
    # e = 23.439 - 0.00000036*d     ; in degree.
    e = 0.4090877233749509 - 6.2831853e-9*d

    # mean anomaly of the Sun, g.
    # g = 357.529 + 0.98560028*d    ; in degree.
    g = 6.2400582213628066 + 0.0172019699945780*d
    g = np.mod(g, twopi)

    # mean longitude of the Sun, q.
    # q = 280.459 + 0.98564736*d   ; in degree.
    q = 4.8949329668507771 + 0.0172027916955899*d
    q = np.mod(q, twopi)

    # geocentric apparent ecliptic longitude, l.
    # l = q + 1.915 sin g + 0.020 sin 2g    ; in degree.
    l = q + 0.0334230551756914*np.sin(g) + 0.0003490658503989*np.sin(2*g)

    # vl - q, mean longitude of the sun.
    # vl = np.mod(279.696678+0.9856473354*dj,360.)/rad
    # q = np.mod(4.881627937990388+0.01720279126623886*dj, twopi)

    # g, mean anomaly of the sun.
    # g = np.mod(358.475845+0.985600267*dj,360.)/rad
    # g = np.mod(6.256583784118852+0.017201969767685215*dj, twopi)

    # slong - l, geocentric apparent ecliptic longitude.
    # slong = (vl + (1.91946-0.004789*t)*sin(g) + 0.020094*sin(2*g))/rad
    # l = q+(0.03350089686033036-2.2884002156881157e-09*dj)*np.sin(g)+0.0003507064598957406*np.sin(2*g)
    # l = np.mod(l, twopi)

    # obliq - e, mean obliquity of the ecliptic.
    # obliq = (23.45229-0.0130125*t)/rad
    # e = 0.40931967763254096-6.217959450123535e-09*dj

    # sin(d) = sin(e) * sin(L)
    sind = np.sin(e)*np.sin(l)
    sdec = np.arcsin(sind)

    # tan(RA) = cos(e)*sin(L)/cos(L)
    srasn = np.arctan2(np.cos(e)*np.sin(l), np.cos(l))
    srasn = np.mod(srasn, twopi)


    # http://aa.usno.navy.mil/faq/docs/GAST.php
    # gst - gmst, greenwich mean sidereal time.
    # gst = np.mod(279.690983+.9856473354*dj+360.*fday+180.,360.)/rad
    # gst = np.mod(4.881528541489487+0.01720279126623886*dj+twopi*fday+np.pi, twopi)
    # gmst = 18.697374558 + 24.06570982441908*d  # in hour
    gmst = 4.894961212735792 + 6.30038809898489*d  # in rad
    gmst = np.mod(gmst, twopi)

    return gmst,l,srasn,sdec,e


def gswgsm(p1,p2,p3, j):
    """
    Converts gsm to gsw coordinates or vice versa.
                       j>0                       j<0
    input:  j,xgsm,ygsm,zgsm           j,xgsw,ygsw,zgsw
    output:    xgsw,ygsw,zgsw           xgsm,ygsm,zgsm

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """
    global e11, e21, e31, e12, e22, e32, e13, e23, e33

    if j > 0:
        xgsw,ygsw,zgsw = [p1,p2,p3]
        xgsm = xgsw*e11 + ygsw*e12 + zgsw*e13
        ygsm = xgsw*e21 + ygsw*e22 + zgsw*e23
        zgsm = xgsw*e31 + ygsw*e32 + zgsw*e33
        return xgsm,ygsm,zgsm
    else:
        xgsm,ygsm,zgsm = [p1,p2,p3]
        xgsw = xgsm*e11 + ygsm*e21 + zgsm*e31
        ygsw = xgsm*e12 + ygsm*e22 + zgsm*e32
        zgsw = xgsm*e13 + ygsm*e23 + zgsm*e33
        return xgsw,ygsw,zgsw


def geomag(p1,p2,p3, j):
    """
    Converts geographic (geo) to dipole (mag) coordinates or vice versa.
                   j>0                       j<0
    input:  j,xgeo,ygeo,zgeo           j,xmag,ymag,zmag
    output:    xmag,ymag,zmag           xgeo,ygeo,zgeo

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """

    #  Attention: subroutine  recalc  must be invoked before geomag in two cases:
    #     /a/  before the first transformation of coordinates
    #     /b/  if the values of time have been changed

    # common /geopack1/ st0,ct0,sl0,cl0,ctcl,stcl,ctsl,stsl,ab(19),bb(8)
    global st0,ct0, sl0,sl0, ctcl,stcl, ctsl,stsl

    if j > 0:
        xgeo,ygeo,zgeo = [p1,p2,p3]
        xmag = xgeo*ctcl+ygeo*ctsl-zgeo*st0
        ymag = ygeo*cl0-xgeo*sl0
        zmag = xgeo*stcl+ygeo*stsl+zgeo*ct0
        return xmag,ymag,zmag
    else:
        xmag,ymag,zmag = [p1,p2,p3]
        xgeo = xmag*ctcl-ymag*sl0+zmag*stcl
        ygeo = xmag*ctsl+ymag*cl0+zmag*stsl
        zgeo = zmag*ct0-xmag*st0
        return xgeo,ygeo,zgeo

def geigeo(p1,p2,p3, j):
    """
    Converts equatorial inertial (gei) to geographical (geo) coords or vice versa.
                   j>0                       j<0
    input:  j,xgei,ygei,zgei           j,xgeo,ygeo,zgeo
    output:    xgeo,ygeo,zgeo           xgei,ygei,zgei

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """

    #  Attention: subroutine  recalc  must be invoked before geomag in two cases:
    #     /a/  before the first transformation of coordinates
    #     /b/  if the values of time have been changed

    # common /geopack1/ a(27),cgst,sgst,b(6)
    global cgst,sgst

    if j > 0:
        xgei,ygei,zgei = [p1,p2,p3]
        xgeo = xgei*cgst+ygei*sgst
        ygeo = ygei*cgst-xgei*sgst
        zgeo = zgei
        return xgeo,ygeo,zgeo
    else:
        xgeo,ygeo,zgeo = [p1,p2,p3]
        xgei = xgeo*cgst-ygeo*sgst
        ygei = ygeo*cgst+xgeo*sgst
        zgei = zgeo
        return xgei,ygei,zgei

def magsm(p1,p2,p3, j):
    """
    Converts dipole (mag) to solar magnetic (sm) coordinates or vice versa
                   j>0                       j<0
    input:  j,xmag,ymag,zmag           j,xsm, ysm, zsm
    output:    xsm, ysm, zsm           xmag,ymag,zmag

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """

    #  Attention: subroutine  recalc  must be invoked before geomag in two cases:
    #     /a/  before the first transformation of coordinates
    #     /b/  if the values of time have been changed

    # common /geopack1/ a(8),sfi,cfi,b(7),ab(10),ba(8)
    global sfi,cfi

    if j > 0:
        xmag,ymag,zmag = [p1,p2,p3]
        xsm = xmag*cfi-ymag*sfi
        ysm = xmag*sfi+ymag*cfi
        zsm = zmag
        return xsm,ysm,zsm
    else:
        xsm,ysm,zsm = [p1,p2,p3]
        xmag = xsm*cfi+ysm*sfi
        ymag = ysm*cfi-xsm*sfi
        zmag = zsm
        return xmag,ymag,zmag

def gsmgse(p1,p2,p3, j):
    """
    converts geocentric solar magnetospheric (gsm) coords to solar ecliptic (gse) ones or vice versa.
                   j>0                       j<0
    input:  j,xgsm,ygsm,zgsm           j,xgse,ygse,zgse
    output:    xgse,ygse,zgse           xgsm,ygsm,zgsm

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """

    # common /geopack1/ a(12),shi,chi,ab(13),ba(8)
    global shi,chi

    if j > 0:
        xgsm,ygsm,zgsm = [p1,p2,p3]
        xgse = xgsm
        ygse = ygsm*chi-zgsm*shi
        zgse = ygsm*shi+zgsm*chi
        return xgse,ygse,zgse
    else:
        xgse,ygse,zgse = [p1,p2,p3]
        xgsm = xgse
        ygsm = ygse*chi+zgse*shi
        zgsm = zgse*chi-ygse*shi
        return xgsm,ygsm,zgsm

def smgsm(p1,p2,p3, j):
    """
    Converts solar magnetic (sm) to geocentric solar magnetospheric (gsm) coordinates or vice versa.
                   j>0                       j<0
    input:  j,xsm, ysm, zsm           j,xgsm,ygsm,zgsm
    output:    xgsm,ygsm,zgsm           xsm, ysm, zsm

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """

    #  Attention: subroutine  recalc  must be invoked before geomag in two cases:
    #     /a/  before the first transformation of coordinates
    #     /b/  if the values of time have been changed

    # common /geopack1/ a(10),sps,cps,b(15),ab(8)
    global sps,cps

    if j > 0:
        xsm,ysm,zsm = [p1,p2,p3]
        xgsm = xsm*cps+zsm*sps
        ygsm = ysm
        zgsm = zsm*cps-xsm*sps
        return xgsm,ygsm,zgsm
    else:
        xgsm,ygsm,zgsm = [p1,p2,p3]
        xsm = xgsm*cps-zgsm*sps
        ysm = ygsm
        zsm = xgsm*sps+zgsm*cps
        return xsm,ysm,zsm

def geogsm(p1,p2,p3, j):
    """
    Converts geographic (geo) to geocentric solar magnetospheric (gsm) coordinates or vice versa.
                   j>0                       j<0
    input:  j,xgeo,ygeo,zgeo           j,xgsm,ygsm,zgsm
    output:    xgsm,ygsm,zgsm           xgeo,ygeo,zgeo

    :param p1,p2,p3: input position
    :param j: flag
    :return: output position
    """

    #  Attention: subroutine  recalc  must be invoked before geomag in two cases:
    #     /a/  before the first transformation of coordinates
    #     /b/  if the values of time have been changed

    # common /geopack1/aa(17),a11,a21,a31,a12,a22,a32,a13,a23,a33,d,b(8)
    global a11,a21,a31,a12,a22,a32,a13,a23,a33

    if j > 0:
        xgeo,ygeo,zgeo = [p1,p2,p3]
        xgsm = a11*xgeo+a12*ygeo+a13*zgeo
        ygsm = a21*xgeo+a22*ygeo+a23*zgeo
        zgsm = a31*xgeo+a32*ygeo+a33*zgeo
        return xgsm,ygsm,zgsm
    else:
        xgsm,ygsm,zgsm = [p1,p2,p3]
        xgeo = a11*xgsm+a21*ygsm+a31*zgsm
        ygeo = a12*xgsm+a22*ygsm+a32*zgsm
        zgeo = a13*xgsm+a23*ygsm+a33*zgsm
        return xgeo,ygeo,zgeo


def geodgeo(p1,p2, j):
    """
    This subroutine (1) converts vertical local height (altitude) h and geodetic
    latitude xmu into geocentric coordinates r and theta (geocentric radial
    distance and colatitude, respectively; also known as ecef coordinates),
    as well as (2) performs the inverse transformation from {r,theta} to {h,xmu}.

    The subroutine uses world geodetic system wgs84 parameters for the earth's
    ellipsoid. the angular quantities (geo colatitude theta and geodetic latitude
    xmu) are in radians, and the distances (geocentric radius r and altitude h
    above the earth's ellipsoid) are in kilometers.

    if j>0, the transformation is made from geodetic to geocentric coordinates using simple direct equations.
    if j<0, the inverse transformation from geocentric to geodetic coordinates is made by means of a fast iterative algorithm.

                  j>0            j<0
    input:    j, h,xmu        j, r,theta
    output:  j, r,theta        j, h,xmu

    Author:  N.A. Tsyganenko
    Date: Dec 5, 2007

    :param h: Altitude in km.
    :param xmu: Geodetic latitude in radian.
    :param r: Geocentric distance in km.
    :param theta: Spherical co-latitude in radian.
    """

    # r_eq is the semi-major axis of the earth's ellipsoid,
    # and beta is its second eccentricity squared
    r_eq, beta = 6378.137, 6.73949674228e-3

    if j>0:     # Direct transformation(GEOD->GEO):
        h,xmu = [p1,p2]
        cosxmu = np.cos(xmu)
        sinxmu = np.sin(xmu)
        den = np.sqrt(cosxmu**2+(sinxmu/(1+beta))**2)
        coslam = cosxmu/den
        sinlam = sinxmu/(den*(1+beta))
        rs = r_eq/np.sqrt(1+beta*sinlam**2)
        x = rs*coslam+h*cosxmu
        z = rs*sinlam+h*sinxmu
        r = np.sqrt(x**2+z**2)
        theta = np.arccos(z/r)
        return r,theta
    else:       # Inverse transformation(GEO->GEOD):
        r,theta = [p1,p2]
        phi = np.pi*0.5-theta
        phi1,dphi,h,xmu,tol = phi,0,0,0,1e-6
        for n in range(100):
            sp = np.sin(phi1)
            arg = sp*(1+beta)/np.sqrt(1+beta*(2+beta)*sp**2)
            xmu = np.arcsin(arg)
            rs = r_eq/np.sqrt(1+beta*np.sin(phi1)**2)
            cosfims = np.cos(phi1-xmu)
            h = np.sqrt((rs*cosfims)**2+r**2-rs**2)-rs*cosfims
            z = rs*np.sin(phi1)+h*np.sin(xmu)
            x = rs*np.cos(phi1)+h*np.cos(xmu)
            rr = np.sqrt(x**2+z**2)
            dphi = np.arcsin(z/rr)-phi
            phi1 -= dphi
            if np.abs(dphi) <= tol: break
        return h,xmu


def sphcar(p1,p2,p3, j):
    """
    Converts spherical coords into cartesian ones and vice versa (theta and phi in radians).
                  j>0            j<0
    input:   j,r,theta,phi     j,x,y,z
    output:      x,y,z        r,theta,phi

    :param r,theta,phi:
    :param x,y,z:
    :param j:
    :return:

    Note: at the poles (x=0 and y=0) we assume phi=0 (when converting from cartesian to spherical coords, i.e., for j<0)
    Last mofification: April 1, 2003 (only some notation changes and more comments added)
    Author:  N.A. Tsyganenko
    """

    if j > 0:
        r,theta,phi = [p1,p2,p3]
        sq=r*np.sin(theta)
        x=sq*np.cos(phi)
        y=sq*np.sin(phi)
        z= r*np.cos(theta)
        return x,y,z
    else:
        x,y,z = [p1,p2,p3]
        sq=x**2+y**2
        r=np.sqrt(sq+z**2)
        if sq != 0:
            sq=np.sqrt(sq)
            phi=np.arctan2(y,x)
            theta=np.arctan2(sq,z)
            if phi < 0: phi += 2*np.pi
        else:
            phi=0.
            if z < 0: theta = np.pi
            else: theta = 0
        return r,theta,phi

def bspcar(theta,phi,br,btheta,bphi):
    """
    Calculates cartesian field components from spherical ones.

    :param theta,phi: spherical angles of the point in radians
    :param br,btheta,bphi: spherical components of the field
    :return: bx,by,bz. cartesian components of the field

    # Last mofification:  April 1, 2003 (only some notation changes and more comments added)
    # Author:  N.A. Tsyganenko
    """

    s= np.sin(theta)
    c= np.cos(theta)
    sf=np.sin(phi)
    cf=np.cos(phi)
    be=br*s+btheta*c

    bx=be*cf-bphi*sf
    by=be*sf+bphi*cf
    bz=br*c-btheta*s

    return bx,by,bz

def bcarsp(x,y,z,bx,by,bz):
    """
    Calculates spherical field components from those in cartesian system

    :param x,y,z: cartesian components of the position vector
    :param bx,by,bz: cartesian components of the field vector
    :return: br,btheta,bphi. spherical components of the field vector

    Last mofification:  April 1, 2003 (only some notation changes and more comments added)
    Author:  N.A. Tsyganenko
    """

    rho2=x**2+y**2
    r=np.sqrt(rho2+z**2)
    rho=np.sqrt(rho2)

    if rho != 0:
        cphi=x/rho
        sphi=y/rho
    else:
        cphi=1.
        sphi=0.

    ct=z/r
    st=rho/r

    br=(x*bx+y*by+z*bz)/r
    btheta=(bx*cphi+by*sphi)*ct-bz*st
    bphi=by*cphi-bx*sphi

    return br,btheta,bphi


def call_external_model(exname, par, ps, x,y,z):
    if exname == 't89':
        return t89.t89(par, ps, x, y, z)
    elif exname == 't96':
        return t96.t96(par, ps, x, y, z)
    elif exname == 't01':
        return t01.t01(par, ps, x, y, z)
    elif exname == 't04':
        return t04.t04(par, ps, x, y, z)
    else:
        raise ValueError

def call_internal_model(inname, x,y,z):
    if inname == 'dipole':
        return dip(x,y,z)
    elif inname == 'igrf':
        return igrf_gsm(x,y,z)
    else:
        raise ValueError

def rhand(x,y,z,parmod,exname,inname):
    """
    Calculates the components of the right hand side vector in the geomagnetic field
    line equation  (a subsidiary subroutine for the subroutine step)

    :param x,y,z:
    :param parmod:
    :param exname: name of the subroutine for the external field.
    :param inname: name of the subroutine for the internal field.

    Last mofification:  March 31, 2003
    Author:  N.A. Tsyganenko
    :return: r1,r2,r3.
    """
    #  common /geopack1/ a(15),psi,aa(10),ds3,bb(8)
    global a, psi, aa, ds3, bb

    bxgsm,bygsm,bzgsm = call_external_model(exname, parmod, psi, x,y,z)
    hxgsm,hygsm,hzgsm = call_internal_model(inname, x,y,z)

    bx=bxgsm+hxgsm
    by=bygsm+hygsm
    bz=bzgsm+hzgsm
    b=ds3/np.sqrt(bx**2+by**2+bz**2)

    r1=bx*b
    r2=by*b
    r3=bz*b

    return r1,r2,r3

def step(x,y,z, ds,errin,parmod,exname,inname):
    """
    Re-calculates {x,y,z}, making a step along a field line.
    model version, the array parmod contains input parameters for that model

    :param x,y,z: the input position
    :param ds: the step size
    :param errin: the permissible error value
    :param parmod: contains input parameters for that model
    :param exname: name of the subroutine for the external field.
    :param inname: name of the subroutine for the internal field.
    :return: x,y,z. The output position

    Last mofification:  March 31, 2003
    Author:  N.A. Tsyganenko
    """

    # common /geopack1/ a(26),ds3,b(8)
    global a, ds3, b

    if errin <=0: raise ValueError
    errcur = errin
    i = 0
    maxloop = 100

    while (errcur >= errin) & (i < maxloop):
        ds3=-ds/3.
        r11,r12,r13 = rhand(x,y,z,parmod,exname,inname)
        r21,r22,r23 = rhand(x+r11,y+r12,z+r13,parmod,exname,inname)
        r31,r32,r33 = rhand(x+.5*(r11+r21),y+.5*(r12+r22),z+.5*(r13+r23),parmod,exname,inname)
        r41,r42,r43 = rhand(x+.375*(r11+3.*r31),y+.375*(r12+3.*r32),z+.375*(r13+3.*r33),parmod,exname,inname)
        r51,r52,r53 = rhand(x+1.5*(r11-3.*r31+4.*r41),y+1.5*(r12-3.*r32+4.*r42),z+1.5*(r13-3.*r33+4.*r43),parmod,exname,inname)
        errcur=np.abs(r11-4.5*r31+4.*r41-.5*r51)+np.abs(r12-4.5*r32+4.*r42-.5*r52)+np.abs(r13-4.5*r33+4.*r43-.5*r53)
        if errcur < errin: break

        ds *= 0.5
        i += 1
    else:
        print('reached maximum loop ...')
        return

    x += 0.5*(r11+4.*r41+r51)
    y += 0.5*(r12+4.*r42+r52)
    z += 0.5*(r13+4.*r43+r53)
    if (errcur < (errin*0.04)) & (np.abs(ds) < 1.33):
        ds *= 1.5

    return x,y,z

def trace(xi,yi,zi,dir,rlim=10,r0=1,parmod=2,exname='t89',inname='igrf',maxloop=1000):
    """
    Traces a field line from an arbitrary point of space to the earth's surface or
    to a model limiting boundary.

    The highest order of spherical harmonics in the main field expansion used
    in the mapping is calculated automatically. if inname=igrf_gsm, then an IGRF model
    field will be used, and if inname=dip, a pure dipole field will be used.

    In any case, before calling trace, one should invoke recalc, to calculate correct
    values of the IGRF coefficients and all quantities needed for transformations
    between coordinate systems involved in this calculations.

    Alternatively, the subroutine recalc can be invoked with the desired values of
    ut sec (to specify the dipole moment), while the values of the dipole
    tilt angle psi (in radians) and its sine (sps) and cosine (cps) can be explicitly
    specified and forwarded to the common block geopack1 (11th, 12th, and 16th elements, resp.)

    :param xi,yi,zi: gsm coords of initial point (in earth radii, 1 re = 6371.2 km). Must be outside the sphere of r0, 
        otherwise tracing will not start.
    :param dir: sign of the tracing direction: if
        dir=1.0 then we move antiparallel to the field vector (e.g. from northern to southern conjugate point), and if
        dir=-1.0 then the tracing goes in the opposite direction.
    :param rlim: upper limit of the geocentric distance, where the tracing is terminated.
    :param r0: radius of a sphere (in re) for which the field line endpoint coordinates xf,yf,zf should be calculated.
    :param parmod: the iopt for T89, or for the other Txx models it is
        a 10-element array containing model parameters, needed for a unique specification of the external field.
        The concrete meaning of the components of parmod depends on a specific version of the external field model.
    :param exname: name of the subroutine for the external field.
    :param inname: name of the subroutine for the internal field.
    :return:
        xf,yf,zf. GSM coords of the last calculated point of a field line
        xx,yy,zz. Arrays containing coords of field line points. Here their maximal length was assumed equal to 999.
        l. Actual number of the calculated field line points. if l exceeds 999, tracing terminates, and a warning is displayed.

    Last mofification:  March 31, 2003
    Author:  N.A. Tsyganenko
    """

    # common /geopack1/ aa(26),dd,bb(8)
    global aa, dd, bb, ds3

    err, l, ds, x,y,z, dd, al = 0.001, 0, 0.5*dir, xi,yi,zi, dir, 0.
    xx = np.array([x])
    yy = np.array([y])
    zz = np.array([z])

    # Here we call RHAND just to find out the sign of the radial component of the field
    # vector, and to determine the initial direction of the tracing (i.e., either away
    # or towards Earth):
    ds3 = -ds/3.
    r1,r2,r3 = rhand(x,y,z,parmod,exname,inname)

    # |ad|=0.01 and its sign follows the rule:
    # (1) if dir=1 (tracing antiparallel to B vector) then the sign of ad is the same as of Br
    # (2) if dir=-1 (tracing parallel to B vector) then the sign of ad is opposite to that of Br
    #     ad is defined in order to initialize the value of rr (radial distance at previous step):
    ad=0.01
    if (x*r1+y*r2+z*r3) < 0: ad=-0.01

    rr=np.sqrt(x**2+y**2+z**2)+ad

    while l < maxloop:
        ryz=y**2+z**2
        r2=x**2+ryz
        r=np.sqrt(r2)
        xr,yr,zr = [x,y,z]

        # check if the line hit the outer tracing boundary; if yes, then terminate the tracing
        if (r >= rlim) | (ryz >= 1600) | (x>=20): break

        # check whether or not the inner tracing boundary was crossed from outside,
        # if yes, then calculate the footpoint position by interpolation
        if (r < r0) & (rr > r): # this is crossing the inner boundary from inside!!!??? looks like it will trace into the boundary and then bounce back out.
            # find the footpoint position by interpolating between the current and previous field line points:
            r1=(r0-r)/(rr-r)
            x=x-(x-xr)*r1
            y=y-(y-yr)*r1
            z=z-(z-zr)*r1
            xr,yr,zr = [x,y,z]
            break

        # check if (i) we are moving outward, or (ii) we are still sufficiently
        # far from Earth (beyond R=5Re); if yes, proceed further:
        if (r >= rr) | (r > 5):
            pass
        # now we moved closer inward (between R=3 and R=5); go to 3 and begin logging
        # previous values of X,Y,Z, to be used in the interpolation (after having
        # crossed the inner tracing boundary):
        else:
            if r >= 3:
                # We entered inside the sphere R=3: to avoid too large steps (and hence inaccurate
                # interpolated position of the footpoint), enforce the progressively smaller
                # stepsize values as we approach the inner boundary R=R0:
                ds = dir
            else:
                fc = 0.2
                if (r-r0) < 0.05: fc = 0.05
                al = fc*(r-r0+0.2)
                ds = dir*al
        rr=r
        x,y,z = step(x,y,z,ds,err,parmod,exname,inname)
        xx = np.append(xx,x)
        yy = np.append(yy,y)
        zz = np.append(zz,z)
        l += 1

    return x,y,z, xx,yy,zz

def shuetal_mgnp(xn_pd,vel,bzimf,xgsm,ygsm,zgsm):
    """
    For any point of space with coordinates (xgsm,ygsm,zgsm) and specified conditions
    in the incoming solar wind, this subroutine:
        (1) determines if the point (xgsm,ygsm,zgsm) lies inside or outside the
            model magnetopause of Shue et al. (jgr-a, v.103, p. 17691, 1998).
        (2) calculates the gsm position of a point {xmgnp,ymgnp,zmgnp}, lying at the model
            magnetopause and asymptotically tending to the nearest boundary point with
            respect to the observation point {xgsm,ygsm,zgsm}, as it approaches the magnetopause.

    :param xn_pd: either solar wind proton number density (per c.c.) (if vel>0)
        or the solar wind ram pressure in nanopascals   (if vel<0)
    :param vel: either solar wind velocity (km/sec)
        or any negative number, which indicates that xn_pd stands
        for the solar wind pressure, rather than for the density
    :param bzimf: imf bz in nanoteslas
    :param xgsm,ygsm,zgsm: gsm position of the observation point in earth radii
    :return: xmgnp,ymgnp,zmgnp. GSM position of the boundary point.
        dist. Distance (in re) between the observation point (xgsm,ygsm,zgsm) and the model magnetopause
        id. position flag: id=+1 (-1) means that the observation point lies inside (outside) of the model magnetopause, respectively.

    Last mofification:  April 4, 2003
    Author:  N.A. Tsyganenko
    """

    # pd is the solar wind dynamic pressure (in npa)
    if vel < 0: pd = xn_pd
    else: pd = 1.94e-6*xn_pd*vel**2

    # Define the angle phi, measured duskward from the noon-midnight meridian plane;
    # if the observation point lies on the x axis, the angle phi cannot be uniquely
    # defined, and we set it at zero:
    if (ygsm != 0) | (zgsm != 0): phi = np.arctan2(ygsm,zgsm)
    else: phi = 0

    # First, find out if the observation point lies inside the Shue et al bdry
    # and set the value of the id flag:
    id = -1
    r0 = (10.22+1.29*np.tanh(0.184*(bzimf+8.14)))*pd**(-0.15151515)
    alpha = (0.58-0.007*bzimf)*(1.+0.024*np.log(pd))
    r = np.sqrt(xgsm**2+ygsm**2+zgsm**2)
    rm = r0*(2./(1.+xgsm/r))**alpha
    if r < rm: id = 1

    #  Now, find the corresponding t96 magnetopause position, to be used as
    #  a starting approximation in the search of a corresponding Shue et al.
    #  boundary point:
    xmt96,ymt96,zmt96,dist,id96 = t96_mgnp(pd,-1.,xgsm,ygsm,zgsm)

    rho2 = ymt96**2+zmt96**2
    r = np.sqrt(rho2+xmt96**2)
    st = np.sqrt(rho2)/r
    ct = xmt96/r

    #  Now, use newton's iterative method to find the nearest point at the Shue et al.'s boundary:
    nit = 0
    while nit < 1000:
        nit += 1

        t = np.arctan2(st,ct)
        rm = r0*(2./(1.+ct))**alpha

        f = r-rm
        gradf_r = 1.
        gradf_t = -alpha/r*rm*st/(1.+ct)
        gradf = np.sqrt(gradf_r**2+gradf_t**2)

        dr = -f/gradf**2
        dt =  dr/r*gradf_t

        r = r+dr
        t = t+dt
        st = np.sin(t)
        ct = np.cos(t)

        ds = np.sqrt(dr**2+(r*dt)**2)
        if ds <= 1.e-4: break
    else: print(' boundary point could not be found; iterations do not converge')


    xmgnp = r*np.cos(t)
    rho =   r*np.sin(t)
    ymgnp = rho*np.sin(phi)
    zmgnp = rho*np.cos(phi)

    dist = np.sqrt((xgsm-xmgnp)**2+(ygsm-ymgnp)**2+(zgsm-zmgnp)**2)

    return xmgnp,ymgnp,zmgnp, dist, id

def t96_mgnp(xn_pd,vel,xgsm,ygsm,zgsm):
    """
    For any point of space with given coordinates (xgsm,ygsm,zgsm), this subroutine defines
    the position of a point (xmgnp,ymgnp,zmgnp) at the T96 model magnetopause, having the
    same value of the ellipsoidal tau-coordinate, and the distance between them. This is
    not the shortest distance d_min to the boundary, but dist asymptotically tends to d_min,
    as the observation point gets closer to the magnetopause.

    The pressure-dependent magnetopause is that used in the t96_01 model
    (Tsyganenko, jgr, v.100, p.5599, 1995; esa sp-389, p.181, oct. 1996)

    :param xn_pd: either solar wind proton number density (per c.c.) (if vel>0)
        or the solar wind ram pressure in nanopascals   (if vel<0)
    :param vel: either solar wind velocity (km/sec)
        or any negative number, which indicates that xn_pd stands
        for the solar wind pressure, rather than for the density
    :param xgsm,ygsm,zgsm: gsm position of the observation point in earth radii
    :return: xmgnp,ymgnp,zmgnp. GSM position of the boundary point.
        dist. Distance (in re) between the observation point (xgsm,ygsm,zgsm) and the model magnetopause
        id. position flag: id=+1 (-1) means that the observation point lies inside (outside) of the model magnetopause, respectively.

    Last mofification:  April 3, 2003
    Author:  N.A. Tsyganenko
    """
    # Define solar wind dynamic pressure (nanopascals, assuming 4% of alpha-particles),
    # if not explicitly specified in the input:
    if vel < 0: pd = xn_pd
    else: pd = 1.94e-6*xn_pd*vel**2

    # ratio of pd to the average pressure, assumed equal to 2 npa:
    # The power index 0.14 in the scaling factor is the best-fit value
    # obtained from data and used in the t96_01 version
    # Values of the magnetopause parameters for  pd = 2 npa:
    rat = pd/2.0
    rat16 = rat**0.14

    a0, s00, x00 = [70.,1.08,5.48]

    # Values of the magnetopause parameters, scaled by the actual pressure:
    # xm is the x-coordinate of the "seam" between the ellipsoid and the cylinder
    # For details on the ellipsoidal coordinates, see the paper:
    # N.A. Tsyganenko, Solution of chapman-ferraro problem for an ellipsoidal magnetopause, planet.space sci., v.37, p.1037, 1989).
    a  = a0/rat16
    s0 = s00
    x0 = x00/rat16
    xm = x0-a


    if (ygsm != 0) | (zgsm != 0): phi = np.arctan2(ygsm,zgsm)
    else: phi = 0

    rho = np.sqrt(ygsm**2+zgsm**2)
    if xgsm < xm:
        xmgnp = xgsm
        rhomgnp = a*np.sqrt(s0**2-1)
        ymgnp = rhomgnp*np.sin(phi)
        zmgnp = rhomgnp*np.cos(phi)
        dist = np.sqrt((xgsm-xmgnp)**2+(ygsm-ymgnp)**2+(zgsm-zmgnp)**2)
        if rhomgnp >  rho: id =  1
        else: id = -1
        return xmgnp,ymgnp,zmgnp, dist, id

    xksi = (xgsm-x0)/a+1.
    xdzt = rho/a
    sq1 = np.sqrt((1.+xksi)**2+xdzt**2)
    sq2 = np.sqrt((1.-xksi)**2+xdzt**2)
    sigma = 0.5*(sq1+sq2)
    tau   = 0.5*(sq1-sq2)

    # Now calculate (x,y,z) for the closest point at the magnetopause
    xmgnp = x0-a*(1.-s0*tau)
    arg = (s0**2-1.)*(1.-tau**2)
    if arg < 0: arg = 0
    rhomgnp = a*np.sqrt(arg)
    ymgnp = rhomgnp*np.sin(phi)
    zmgnp = rhomgnp*np.cos(phi)

    # Now calculate the distance between the points {xgsm,ygsm,zgsm} and {xmgnp,ymgnp,zmgnp}:
    # (in general, this is not the shortest distance d_min, but dist asymptotically tends
    # to d_min, as we are getting closer to the magnetopause):
    dist = np.sqrt((xgsm-xmgnp)**2+(ygsm-ymgnp)**2+(zgsm-zmgnp)**2)
    if sigma > s0: id = -1  # id = -1 means that the point lies outside
    else: id = 1            # id =  1 means that the point lies inside

    return xmgnp,ymgnp,zmgnp, dist, id


init_igrf()





def sun_position(doy, time, glat, glon, height, **kwargs):
    


subroutine soco (ld,t,flat,Elon,height,
     &          DECLIN, ZENITH, SUNRSE, SUNSET)
c--------------------------------------------------------------------
c       s/r to calculate the solar declination, zenith angle, and
c       sunrise & sunset times  - based on Newbern Smith's algorithm
c       [leo mcnamara, 1-sep-86, last modified 16-jun-87]
c       {dieter bilitza, 30-oct-89, modified for IRI application}
c
c in:   ld      local day of year
c       t       local hour (decimal)
c       flat    northern latitude in degrees
c       elon    east longitude in degrees
c		height	height in km
c
c out:  declin      declination of the sun in degrees
c       zenith      zenith angle of the sun in degrees
c       sunrse      local time of sunrise in hours 
c       sunset      local time of sunset in hours 
c-------------------------------------------------------------------
c
        common/const/dtr,pi     /const1/humr,dumr
c amplitudes of Fourier coefficients  --  1955 epoch.................
        data    p1,p2,p3,p4,p6 /
     &  0.017203534,0.034407068,0.051610602,0.068814136,0.103221204 /
c
c s/r is formulated in terms of WEST longitude.......................
        wlon = 360. - Elon
c
c time of equinox for 1980...........................................
        td = ld + (t + Wlon/15.) / 24.
        te = td + 0.9369
c
c declination of the sun..............................................
        dcl = 23.256 * sin(p1*(te-82.242)) + 0.381 * sin(p2*(te-44.855))
     &      + 0.167 * sin(p3*(te-23.355)) - 0.013 * sin(p4*(te+11.97))
     &      + 0.011 * sin(p6*(te-10.41)) + 0.339137
        DECLIN = dcl
        dc = dcl * dtr
c
c the equation of time................................................
        tf = te - 0.5
        eqt = -7.38*sin(p1*(tf-4.)) - 9.87*sin(p2*(tf+9.))
     &      + 0.27*sin(p3*(tf-53.)) - 0.2*cos(p4*(tf-17.))
        et = eqt * dtr / 4.
c
        fa = flat * dtr
        phi = humr * ( t - 12.) + et
c
        a = sin(fa) * sin(dc)
        b = cos(fa) * cos(dc)
        	cosx = a + b * cos(phi)
        	if(abs(cosx).gt.1.) cosx=sign(1.,cosx)
        zenith = acos(cosx) / dtr
c
c calculate sunrise and sunset times --  at the ground...........
c see Explanatory Supplement to the Ephemeris (1961) pg 401......
c sunrise at height h metres is at...............................
		h=height*1000.
        chih = 90.83 + 0.0347 * sqrt(h)
c this includes corrections for horizontal refraction and........
c semi-diameter of the solar disk................................
        ch = cos(chih * dtr)
        cosphi = (ch -a ) / b
c if abs(secphi) > 1., sun does not rise/set.....................
c allow for sun never setting - high latitude summer.............
        secphi = 999999.
        if(cosphi.ne.0.) secphi = 1./cosphi
        sunset = 99.
        sunrse = 99.
        if(secphi.gt.-1.0.and.secphi.le.0.) return
c allow for sun never rising - high latitude winter..............
        sunset = -99.
        sunrse = -99.
        if(secphi.gt.0.0.and.secphi.lt.1.) return
c
        	cosx = cosphi
        	if(abs(cosx).gt.1.) cosx=sign(1.,cosx)
        phi = acos(cosx)
        et = et / humr
        phi = phi / humr
        sunrse = 12. - phi - et
        sunset = 12. + phi - et
        if(sunrse.lt.0.) sunrse = sunrse + 24.
        if(sunset.ge.24.) sunset = sunset - 24.
c special case sunrse > sunset
        if(sunrse.gt.sunset) then
        	sunx=sign(99.0,flat)
        	if(ld.gt.91.and.ld.lt.273) then
        		sunset = sunx
        		sunrse = sunx
        	else
        		sunset = -sunx
        		sunrse = -sunx
        	endif
        	endif
c
        return
        end
c




def func_1(a, b):
    print(a)
    print(a+b)


def func_2(a, b, c):
    func_1(a, b)
    print(c)


class B:
    def __init__(self, value):
        self.c = value
        self.d = value * 2

    def test(self, value):
        self.c = value*3


class A(B):
    def __init__(self, value):
        self.var = None
        self.var = self.test(value)

    @property
    def var(self):
        return self._var

    @var.setter
    def var(self, value):
        if value is None:
            self._var = None
        else:
            self._var = B(value)




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import numpy as np
from numpy.polynomial import Polynomial
from scipy.interpolate import interp1d


def trig_arctan_to_sph_lon(x, y):
    isscalar = False
    if np.isscalar(x) and np.isscalar(y):
        isscalar = True
        x = np.array([x])
        y = np.array([y])
    old_settings = np.seterr(divide='ignore')     
    angle = np.arctan(np.divide(y, x))
    angle = np.where(x<0, angle + np.pi, angle)
    angle = np.mod(angle, 2*np.pi)
    np.seterr(**old_settings)
    if isscalar:
        angle = angle[0]
    return angle


def calc_curve_tangent_slope(x, y, degree=3, unit='radians'):
    poly = Polynomial.fit(x, y, degree)
    c = poly.convert().coef
    c_deriv = [i * cd for i, cd in enumerate(c)]
    c_deriv = c[1:]
    f_deriv = Polynomial(c_deriv)
    tan_slope = f_deriv(x)

    slope = np.arctan(tan_slope)

    if unit == 'degree':
        slope = slope / np.pi * 180.

    return slope


def interp_period_data(x, y, xq,  period=360., method='linear', **kwargs):
    factor = 2 * np.pi / period

    sin_y = np.sin(y * factor)
    cos_y = np.cos(y * factor)
    itpf_sin = interp1d(x, sin_y, kind=method, **kwargs)
    itpf_cos = interp1d(x, cos_y, kind=method, **kwargs)
    sin_y_i = itpf_sin(xq)
    cos_y_i = itpf_cos(xq)
    rad = np.sign(sin_y_i) * (np.pi / 2 - np.arcsin(cos_y_i))
    rad = np.mod(rad, 2*np.pi)
    yq = rad / factor
    return yq


if __name__ == "__main__":
    x1 = np.array(range(10))
    y1 = np.array(range(10)) * 5
    calc_curve_tangent_slope(x1, y1)



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from geospacelab import preferences as prf

from geospacelab.datahub.sources.esa_eo.swarm.downloader import Downloader as DownloaderModel


class Downloader(DownloaderModel):

    def __init__(
            self, dt_fr, dt_to,
            sat_id=None,
            data_type='TCT02',
            file_version=None,
            data_file_root_dir=None,
            ftp_data_dir=None,
            force=True, direct_download=True, **kwargs
    ):

        self.sat_id = sat_id

        if data_type == 'TCT02':
            frq_str = '2Hz'
        elif data_type == 'TCT16':
            frq_str = '16Hz'
        else:
            raise NotImplementedError

        if ftp_data_dir is None:
            ftp_data_dir = f'Advanced/Plasma_Data/{frq_str}_TII_Cross-track_Dataset/Sat_{sat_id.upper()}'

        if data_file_root_dir is None:
            data_file_root_dir = prf.datahub_data_root_dir / "ESA" / "SWARM" / "Advanced" / "EFI-TII" / data_type

        super(Downloader, self).__init__(
            dt_fr, dt_to,
            data_file_root_dir=data_file_root_dir,
            ftp_data_dir=ftp_data_dir,
            file_version=file_version,
            force=force, direct_download=direct_download, **kwargs
        )

    def download(self, **kwargs):

        done = super(Downloader, self).download(**kwargs)
        return done

    def search_files(self, **kwargs):

        file_list, versions = super(Downloader, self).search_files(**kwargs)

        return file_list, versions
        # version control


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import pathlib
import geospacelab.toolbox.utilities.pylogging as mylog


def list_eiscat_hdf5_variables(fh5, var_groups=None, var_names_queried=None, display=False):
    if isinstance(fh5, str) or isinstance(fh5, pathlib.PurePath):
        fh5 = h5py.File(fh5, 'r')
    elif not isinstance(fh5, h5py._hl.files.File):
        raise TypeError
    if var_groups is None:
        var_groups = ['par0d', 'par1d', 'par2d']
    var_info = {
        'name': [],
        'unit': [],
        'index': [],
        'group': [],
        'name_GUISDAP': [],
        'note': []
    }
    for var_group in var_groups:
        metadata_var = fh5['metadata'][var_group]
        for ind in range(metadata_var.shape[0]):
            var_name = metadata_var[ind, 0].decode('UTF-8').strip()

            if var_names_queried is not None:
                if var_name not in var_names_queried:
                    continue
            var_note = metadata_var[ind, 1].decode('UTF-8').strip()
            var_unit = metadata_var[ind, 2].decode('UTF-8').strip()
            var_name_GUISDAP = metadata_var[ind, 3].decode('UTF-8').strip()

            var_info['name'].append(var_name)
            var_info['index'].append(ind)
            var_info['unit'].append(var_unit)
            var_info['note'].append(var_note)
            var_info['group'].append(var_group)
            var_info['name_GUISDAP'].append(var_name_GUISDAP)

    len_vars = len(var_info['name'])
    if len_vars == 0:
        print('Cannot find the queried variable!')

    if display:
        mylog.simpleinfo.info("Listing the variables in the ESICAT hdf5 file ({})".format(fh5))
        mylog.simpleinfo.info(
            '{:<4s}{:20s}{:10s}{:10s}{:20s}{:20s}{:^60s}'.format(
                '', 'Name', 'Group', 'Index', 'Unit', 'Name (GUISDAP)', 'Note'
            )
        )
        for i in range(len_vars):
            mylog.simpleinfo.info('{:<4s}{:20s}{:10s}{:<10d}{:20s}{:20s}{:60s}'.format(
                '', var_info['name'][i], var_info['group'][i], var_info['index'][i],
                var_info['name_GUISDAP'][i], var_info['unit'][i], var_info['note'][i])
            )
        print()
    return var_info






# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from geospacelab.datahub import VariableModel as Var
import geospacelab.visualization.mpl.colormaps as cm

import numpy as np

database = 'ESA/EarthOnline'

timestamps = {
    'SC_DATETIME': 'SC_DATETIME',
}


default_colormap = "gist_ncar"

default_plot_config = {
    'line':         {
        'linestyle':        '-',
        'linewidth':        1.5,
        'marker':           '',
        'markersize':       3,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
    }
}

configured_variables = {}
visual = 'on'

depend_0 = {'UT': 'SC_DATETIME',
            'GEO_LAT': 'SC_GEO_LAT', 'GEO_LON': 'SC_GEO_LON',
            'AACGM_LAT': 'SC_AACGM_LAT', 'AACGM_LON': 'SC_AACGM_LON', 'AACGM_MLT': 'SC_AACGM_MLT'}
# depend_c = {'SPECTRA': 'EMISSION_SPECTRA'}

####################################################################################################################
var_name = 'n_e'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Plasma density from ion current'
var.label = r'$n_e$'
var.unit = 'cm-3'
var.unit_label = r'cm$^{-3}$'
var.group = r'$n_e$'
# var.error = var_name + '_err'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
# axis[1].lim = [np.nan, np.nan]
axis[1].label = '@v.group'
axis[1].unit = ''
axis[2].label = '@v.label'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'T_e_HGN'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electron temperature from the high gain probe'
var.label = r'$T_e^{H}$'
var.unit = 'K'
var.group = r'$T$'
# var.error = var_name + '_err'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
# axis[1].lim = [-2000, 2000]
axis[1].label = '@v.group'
axis[2].label = '@v.label'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'T_e_LGN'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electron temperature from the low gain probe'
var.label = r'$T_e^{L}$'
var.unit = 'K'
var.group = r'$T$'
# var.error = var_name + '_err'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
# axis[1].lim = [-2000, 2000]
axis[1].label = '@v.group'
axis[2].label = '@v.label'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'T_e'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electron temperature blended'
var.label = r'$T_e$'
var.unit = 'K'
var.group = r'$T$'
# var.error = var_name + '_err'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [0, 10000]
axis[1].label = '@v.group'
axis[2].label = '@v.label'
axis[1].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'QUALITY_FLAG'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Flag bits indicating data quality or properties'
var.label = r'Flag'
var.unit = ''
var.group = r''
# var.error = var_name + '_err'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
# axis[1].lim = [-2000, 2000]
axis[1].label = '@v.group'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLAB"
__credits__ = ["Lei Cai"]
__license__ = "BSD-3-Clause License"
__version__ = "0.6.0"
__maintainer__ = "Lei Cai"
__email__ = "lei.cai@oulu.fi"
__status__ = "Developing"
__docformat__ = "reStructureText"


import geospacelab.config as config

preferences = config.Preferences()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as dtparse
import numpy as np
import re
import requests
import bs4
import os
import pathlib

from geospacelab import preferences as pfr
import geospacelab.datahub.sources.madrigal as madrigal
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool


def test():
    dt_fr = datetime.datetime(2014, 3, 10)
    dt_to = datetime.datetime(2014, 3, 10)
    download_obj = Downloader(dt_fr, dt_to, sat_id='f16', file_type='s1')


data_type_dict = {
    's1': 'ion drift',
    's4': 'plasma temp',
    'e': 'flux/energy',
    'ssies2': 'SSIES-2',
    'hp':     'Hemispherical power'
}


class Downloader(object):
    """Download the GNSS TEC data
    """

    def __init__(self, dt_fr, dt_to, sat_id=None, file_type=None, data_file_root_dir=None, force=False,
                 user_fullname=madrigal.default_user_fullname,
                 user_email=madrigal.default_user_email,
                 user_affiliation=madrigal.default_user_affiliation):
        self.user_fullname = user_fullname
        self.user_email = user_email
        self.user_affiliation = user_affiliation

        self.instrument_code = 8100

        dt_fr = dttool.get_start_of_the_day(dt_fr)
        dt_to = dttool.get_end_of_the_day(dt_to)

        if sat_id is None:
            sat_id = input("input the sat_id (e.g., f16)")
        self.sat_id = sat_id

        if file_type is None:
            file_type = input("Input the file type (s1, s4, or e):")
        self.file_type = file_type
        if dt_fr == dt_to:
            dt_to = dt_to + datetime.timedelta(hours=23, minutes=59)
        self.dt_fr = dt_fr  # datetime from
        self.dt_to = dt_to  # datetime to
        self.force = force

        if data_file_root_dir is None:
            self.data_file_root_dir = pfr.datahub_data_root_dir / 'Madrigal' / 'DMSP'
        else:
            self.data_file_root_dir = data_file_root_dir
        self.done = False

        self.madrigal_url = "http://cedar.openmadrigal.org/"
        self.download_madrigal_files()

    def download_madrigal_files(self, download_pp=False):

        mylog.simpleinfo.info("Searching data from the Madrigal database ...")
        exp_list, _, database = madrigal.utilities.list_experiments(
            self.instrument_code, self.dt_fr, self.dt_to, madrigal_url=self.madrigal_url
        )
        for exp in exp_list:
            files = database.getExperimentFiles(exp.id)
            for file in files:
                if self.file_type == 'hp':
                    rpattern = data_type_dict[self.file_type]
                else:
                    rpattern = self.sat_id.upper() + '.*' + data_type_dict[self.file_type]

                m = re.search(rpattern, file.kindatdesc)
                if m is None:
                    continue

                file_path_remote = pathlib.Path(file.name)
                file_name = file_path_remote.name
                m = re.search(r'([\d]{8})', file_name)
                dtstr = m.group()
                thisday = datetime.datetime.strptime(dtstr, "%Y%m%d")
                if thisday < datetime.datetime(self.dt_fr.year, self.dt_fr.month, self.dt_fr.day):
                    continue
                if thisday > datetime.datetime(self.dt_to.year, self.dt_to.month, self.dt_to.day):
                    continue

                data_file_dir = self.data_file_root_dir / thisday.strftime("%Y%m") / thisday.strftime('%Y%m%d')
                data_file_dir.mkdir(parents=True, exist_ok=True)

                data_file_path = data_file_dir / file_name
                if data_file_path.is_file():
                    mylog.simpleinfo.info("The file {} has been downloaded.".format(data_file_path.name))
                    if not self.force:
                        continue
                    else:
                        print('Force downloading ...')

                mylog.simpleinfo.info("Downloading  {} from the Madrigal database ...".format(file_name))
                database.downloadFile(
                    file_path_remote, data_file_path,
                    self.user_fullname, self.user_email, self.user_affiliation,
                    "hdf5"
                )
                self.done = True
                mylog.simpleinfo.info("Save the file as {} in the directory {}".format(file_name, data_file_dir))


if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import cftime
import re
import datetime

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool

pulse_code_dict = {
    'alternating': 97,
    'barker': 98,
    'single pulse': 115,
}

antenna_code_dict = {
    'zenith': 32,
    'misa': 31,
}

var_name_dict = {
    'n_pp_log': 'popl',
    'n_pp_log_err': 'dpopl',
    'n_e': 'ne',
    'n_e_err': 'dne',
    'v_i_UP': 'vi3',
    'v_i_UP_err': 'dvi3',
    'T_i': 'ti',
    'T_i_err': 'dti',
    'T_r': 'tr',
    'T_r_err': 'dtr',
    'comp_mix': 'pm',
    'comp_mix_err': 'pm',
    'coef_T_i_T_r': 'cctitr',
    'coef_T_i_T_r_err': 'dcctitr',
    'GEO_LAT': 'gdlat',
    'GEO_LON': 'glon',
    'TEC': 'tec',
    'n_e_max': 'nemax',
    'h_max': 'hmax',
    'GEO_ALT': 'gdalt',
}


class Loader:
    """
    :param file_path: the file's full path
    :type file_path: pathlib.Path object
    :param file_type: the specific file type for the file being loaded. Options: ['TEC-MAT'], 'TEC-LOS', 'TEC-sites')
    :type file_type: str
    :param load_data: True, load without calling the method "load_data" separately.
    :type load_data: bool
    """
    def __init__(self, file_path, load_data=True):
        self.file_path = file_path
        self.variables = {}
        self.metadata = {}

        self.done = False
        if load_data:
            self.load()

    def load(self):
        variables = {}
        with h5py.File(self.file_path, 'r') as fh5:
            data_fh5 = fh5['Data']
            vars_fh5 = {}
            fh5_vars_1d = data_fh5['Array Layout']['1D Parameters']
            for var_name in fh5_vars_1d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_1d[var_name])[:, np.newaxis]
            fh5_vars_2d = data_fh5['Array Layout']['2D Parameters']
            for var_name in fh5_vars_2d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_2d[var_name]).T
            vars_fh5['gdalt'] = np.array(data_fh5['Array Layout']['gdalt'])[np.newaxis, :]
            vars_fh5['timestamps'] = np.array(data_fh5['Array Layout']['timestamps'])[:, np.newaxis]
            for var_name, var_name_fh5 in var_name_dict.items():
                if var_name_fh5 not in vars_fh5.keys():
                    mylog.StreamLogger.warning(
                        f"The requested variable {var_name_fh5} does not exist in the data file!")
                    variables[var_name] = None
                    continue
                variables[var_name] = vars_fh5[var_name_fh5]
            variables['DATETIME'] = dttool.convert_unix_time_to_datetime_cftime(vars_fh5['timestamps'])
            variables['HEIGHT'] = np.tile(variables['GEO_ALT'], (variables['DATETIME'].shape[0], 1))
            variables['T_e'] = variables['T_i'] * variables['T_r']
            variables['T_e_err'] = variables['T_e'] * np.sqrt((variables['T_i_err'] / variables['T_i']) ** 2
                                                              + (variables['T_r_err'] / variables['T_r']) ** 2)
        self.variables = variables


if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path("/home/lei/afys-data/Madrigal/Millstone_ISR/2016/20160314/Millstone_ISR_combined_20160314.005.hdf5")
    Loader(file_path=fp)